/**
 * File:        ECE 4760 Final Project FA18
 * Author:      Adam Macioszek and Felipe Fortuna
 * Target PIC:  PIC32MX250F128B
 */

////////////////////////////////////
// includes
#include "config_1_2_3.h"
#include "pt_cornell_1_2_3.h"
#include "tft_master.h"
#include "tft_gfx.h"
#include "i2c_lib.h"
#include "kalman_lib.h"
#include <math.h>
#include <stdlib.h>


// === global variables ========================================================
// thread structures
static struct pt pt_debug, pt_imu, pt_user;
static char buffer[60]; // for printing
static char buffer1[60]; // for printing

// IMU readings
static float roll;
static float kal_roll;  // Kalman filter roll output (global for print)
static float kal_dtfl;

// PID Loop
volatile float balance_pt ;     // IMU reading to aim for.
static   float Kp;                    // control loop proportional gain
static   float Ki;                    // control loop integral gain
static   float Kd;                    // control loop derivative gain
static   float error_angle;           // proportional term, equal to kal_roll
static   float last_error = 0.0;      // Store previous error value for differential term
volatile float diff_ctrl;             // differential term
volatile float intg_ctrl = 0.0;       // integral term, must be static
static   int   PID_out;

// PWM parameters:
static int gen_period = 2000;    // period of PWM in t2 ticks. PWM frequency = t2_frequency (40MHz) / gen_period (2000) = 20usuvc/c kHz;
static int duty_cycle;             // Duty cycle of PWM (global for print)
static   char  direction;             // 0 means forwards, 1 means backwards (global for print)
volatile char  prev_direction;      // Utility: don't set direction pins unless direction changed

// Function declaration
void change_direction(char dir);

 // === Timer 4 ISR (for PWM) ============================================                                          
void __ISR(_TIMER_4_VECTOR, ipl1) Timer4Handler(void) {
    // PID Loop
    error_angle = kal_roll; 
    diff_ctrl = (error_angle - last_error);   
    intg_ctrl = (error_angle*last_error > 0) ? (intg_ctrl + error_angle) : 0;
    PID_out =  (Kp * error_angle) + (Kd * diff_ctrl) + (Ki * intg_ctrl);
    last_error = kal_roll;

    // Take abs of PID_out, cutoff at {gen_period}, and set duty cycle
    
    // Set direction of motors   
    if (kal_roll < 0) direction = 0;
    else              direction = 1;
    
    if (PID_out < 0 ) PID_out *= -1;
    if (PID_out > gen_period) PID_out = gen_period;
    duty_cycle = gen_period- PID_out ; // duty cycle is reverse of PID_out PWM, since motor is active low
    
    if (prev_direction != direction) change_direction(direction);
    prev_direction = direction;

    SetDCOC3PWM(duty_cycle); 
    mT4ClearIntFlag();
} // Timer 4 ISR


// === Debug Thread ============================================================
static PT_THREAD (protothread_debug(struct pt *pt))
{
    PT_BEGIN(pt);
    tft_setTextColor(ILI9340_WHITE);
    tft_setTextSize(2);
    
    while(1) {
        
        tft_fillRect(100, 0, 140, 18, ILI9340_BLUE);
        sprintf(buffer, "roll   = %.6f", roll);//roll.6f
        PT_YIELD(pt);
        tft_setCursor(0, 0);
        tft_writeString(buffer);
        PT_YIELD(pt);

        tft_fillRect(100, 20, 140, 18, ILI9340_BLUE);
        sprintf(buffer, "kRoll  = %.6f", kal_roll);
        PT_YIELD(pt);
        tft_setCursor(0, 20);
        tft_writeString(buffer);
        PT_YIELD(pt);

        tft_fillRect(100, 60, 140, 18, ILI9340_BLUE);
        sprintf(buffer, "P      = %.2f", Kp);
        PT_YIELD(pt);
        tft_setCursor(0, 60);
        tft_writeString(buffer);
        PT_YIELD(pt);

        tft_fillRect(100, 80, 140, 18, ILI9340_BLUE);
        sprintf(buffer, "I      = %.4f", Ki);
        PT_YIELD(pt);
        tft_setCursor(0, 80);
        tft_writeString(buffer);
        PT_YIELD(pt);
        
        tft_fillRect(100, 100, 140, 18, ILI9340_BLUE);
        sprintf(buffer, "D    =   %.4f", Kd);
        PT_YIELD(pt);
        tft_setCursor(0, 100);
        tft_writeString(buffer);
        PT_YIELD(pt);

        tft_fillRect(100, 140, 140, 18, ILI9340_BLUE);
        sprintf(buffer, "center = %.4f", balance_pt);
        PT_YIELD(pt);
        tft_setCursor(0, 140);
        tft_writeString(buffer);
        PT_YIELD(pt);

        tft_fillRect(100, 160, 140, 18, ILI9340_BLUE);
        sprintf(buffer, "imu_dt = %.4f ms", kal_dtfl * 1000.0);
        PT_YIELD(pt);
        tft_setCursor(0, 160);
        //sprintf(buffer, "dt     = %d ms", pid_dt);
        tft_writeString(buffer);
        PT_YIELD(pt);

        tft_fillRect(100, 200, 140, 18, ILI9340_BLUE);
        sprintf(buffer, "PWM    = %.3f", (float)(gen_period - duty_cycle)/(float)(gen_period));
        PT_YIELD(pt);
        tft_setCursor(0, 200);
        tft_writeString(buffer);
        PT_YIELD(pt);

        tft_fillRect(100, 220, 140, 18, ILI9340_BLUE);
        sprintf(buffer, "desRPM = %d", (int) (140*((float)(gen_period-duty_cycle)/(float)(gen_period))));
        PT_YIELD(pt);
        tft_setCursor(0, 220);
        tft_writeString(buffer);
        PT_YIELD(pt);

    } // END WHILE(1)

  PT_END(pt);
} // Debug Thread


// === IMU Thread ==============================================================
static PT_THREAD (protothread_imu(struct pt *pt))
{
  PT_BEGIN(pt);
  PT_YIELD_TIME_msec(1000); // wait a second before calibrating Gyroscope
  calibrateGyroscope();
  
  
  // IMU structures
  int acc_raw[3]; 
  int gyr_raw[3];
  float gyroYrate;
  static int kal_t0;                  // Utility: store previous ReadTimer1
  kal_t0 = ReadTimer1();
  static int kal_dt;
 
  float kal_temp;
  while(1) { 
    // Timing Calculations
    kal_temp = ReadTimer1();
    kal_dt = (kal_temp < kal_t0) ? (0xffff - kal_t0 + kal_temp) : (kal_temp - kal_t0); // in t1 periods
    kal_dtfl = ((float) (kal_dt))/(625000.0);
            //(40000000.0); // pid_dt in seconds
    kal_t0 = kal_temp;  
      
    // Get IMU readings
    readIMU_acc(acc_raw);
    //readIMU_gyr(gyr_raw);
    gyr_raw[1] = readIMU_gyrY();
       
    // Calculate roll
    roll = 57.2958 * atan2(-acc_raw[0], acc_raw[2]);
    roll = roll - balance_pt; // Set 0 roll to balance point
    gyroYrate = gyr_raw[1] / 131.0; // in deg/sec
    
    kal_roll = 0.99 *(kal_roll + gyroYrate*(kal_dtfl)) + 0.01*(roll); //a *(kal_roll + gyroYrate*(dt)*0.5) +(1-a)*(roll);

    PT_YIELD(pt);
  } // END WHILE(1)

  PT_END(pt);
} // IMU Thread

// === User Input Thread =======================================================
static PT_THREAD (protothread_user(struct pt *pt))
{
    PT_BEGIN(pt);
    while(1) {
        // Read ADC
        int potVal1 = ReadADC10(0);
        Kp = 600 * (potVal1/1024.0);
        
        int potVal2 = ReadADC10(2);
        balance_pt = -1*(10 * (potVal2/1024.0)+85);

        int potVal3 = ReadADC10(1);   
        Ki = 2 * (potVal3/1024.0);
        PT_YIELD_TIME_msec(100) ; 
      } // END WHILE(1)

  PT_END(pt);
} // User Input Thread

// === Main  ===================================================================
int main(void)
{
  // === configure peripherals ===============================
  /* PORT Description:
   * RA0:  Pot 1 (sets P)
   * RA1:  Pot 3 (sets I)
   * RB3:  Pot 2 (sets D)
   *
   * RB0:  LCD
   * RB1:  LCD
   * RB2:  LCD 
   * RB11: LCD
   * RB14: LCD
   * 
   * RB9:  IMU SDA line
   * RB8:  IMU SCL line
   * 
   * RB7:  Left  Motor yellow (direction)
   * RB5:  Right Motor yellow (direction)
   * RB13: Left  Motor green  (feedback)
   * RB15: Right Motor green  (feedback) (unused!)
   * RA3:  Both  Motor blue   (PWM)
   */  
    
  // Configure Motor I/O pins
  mPORTASetPinsDigitalOut(BIT_4 | BIT_2); // PWM
  mPORTBSetPinsDigitalOut(BIT_5 | BIT_7); // Direction (use drive_forward() or drive_back() functions)
  
  
  // Configure Timer 4 to run PID loop
  int n = 50;   
  OpenTimer4(T4_ON | T4_SOURCE_INT | T4_PS_1_32, 6250); // use timer 3 for input capture, set 1:32 prescaler, so it's running at 1.25MHz, interrupts every 10ms 12500
  ConfigIntTimer4(T4_INT_ON | T4_INT_PRIOR_1); 
  mT4ClearIntFlag(); 
  
  // Configure Output Compare 3 and Timer 2 for PWM
  OpenTimer2(T2_ON | T2_SOURCE_INT | T2_PS_1_1, gen_period); // Timer 2 at 40MHz, interrupts at 2000
  //ConfigIntTimer2(T2_INT_ON | T2_INT_PRIOR_1);
  //mT2ClearIntFlag();
  OpenOC3(OC_ON | OC_TIMER2_SRC | OC_PWM_FAULT_PIN_DISABLE, 1000, 1000); //
  PPSOutput(4, RPA3, OC3); // connect A3 to OC3 PWM 

  // Configure timer 1 to measure time between IMU readings, used by Kalman filter
  OpenTimer1(T1_ON | T1_SOURCE_INT | T1_PS_1_64, 0xffff); // Timer 1 to time IMU reading at f=625kHz

  
  // Configure timer 4 for running PWM between i2c readings
  //OpenTimer1(T1_ON | T1_SOURCE_INT | T1_PS_1_8, 0xffff); // Timer 1 at 5MHz
  //ConfigIntTimer1(T1_INT_ON | T1_INT_PRIOR_2);
  //mT2ClearIntFlag();
  
  // Configure ADC 
  ANSELA = 0; ANSELB = 0; 
  CloseADC10(); // ensure the ADC is off before setting the configuration
  #define PARAM1  ADC_FORMAT_INTG16 | ADC_CLK_AUTO | ADC_AUTO_SAMPLING_ON //
  #define PARAM2  ADC_VREF_AVDD_AVSS | ADC_OFFSET_CAL_DISABLE | ADC_SCAN_ON | ADC_SAMPLES_PER_INT_3 | ADC_ALT_BUF_OFF | ADC_ALT_INPUT_OFF
  #define PARAM3  ADC_CONV_CLK_PB | ADC_SAMPLE_TIME_15 | ADC_CONV_CLK_Tcy 
  #define PARAM4  ENABLE_AN0_ANA | ENABLE_AN5_ANA |  ENABLE_AN1_ANA // one for each potentiometer
  #define PARAM5  SKIP_SCAN_AN2 | SKIP_SCAN_AN3 | SKIP_SCAN_AN4 | SKIP_SCAN_AN6 | SKIP_SCAN_AN7 | SKIP_SCAN_AN8 | SKIP_SCAN_AN9 | SKIP_SCAN_AN10 | SKIP_SCAN_AN11 | SKIP_SCAN_AN12 | SKIP_SCAN_AN13 | SKIP_SCAN_AN14 | SKIP_SCAN_AN15
  SetChanADC10( ADC_CH0_NEG_SAMPLEA_NVREF); // 
  OpenADC10( PARAM1, PARAM2, PARAM3, PARAM4, PARAM5 ); // configure ADC using the parameters defined above
  EnableADC10(); // Enable the ADC

  // configure MPU 6050 
  OpenI2C1( I2C_ON, 48 ); // Enable I2C, baud rate generator BRG = (Fpb / 2 / baudrate) - 2
  i2c_write(0x6b, 0x00); // power on
  i2c_write(0x19, 0x07); // set sample rate to 1kHz
  i2c_write(0x1a, 0x00); // disable fsync, set 260Hz Acc filter and 256Hz Gyro filter
  i2c_write(0x1b, 0x00); // Set the gyro full-scale  range to +-250 degrees/second
  i2c_write(0x1c, 0x00); // Set the acc  full-scale range to +-2g
  
  // === configure threads + lcd ==============================
  INTEnableSystemMultiVectoredInt();
  PT_setup();
  
  // TFT Display
  tft_init_hw();
  tft_begin();
  tft_fillScreen(ILI9340_BLACK);
  tft_setRotation(2);
  tft_setTextColor(ILI9340_WHITE);
  tft_setTextSize(2);
   
  // init threads
  PT_INIT(&pt_imu);
  PT_INIT(&pt_user);
  //PT_INIT(&pt_debug);
  while(1) {
      PT_SCHEDULE(protothread_imu(&pt_imu));
      PT_SCHEDULE(protothread_user(&pt_user));
      PT_SCHEDULE(protothread_debug(&pt_debug));
  }
} // main

void change_direction(char dir){
    if (dir){
        PORTSetBits(IOPORT_B, BIT_7);
        PORTClearBits(IOPORT_B, BIT_5);
    }
    else{
        PORTSetBits(IOPORT_B, BIT_5);
        PORTClearBits(IOPORT_B, BIT_7);
    }
}

// Print debug info
void print_debug(){  
    
    
}